SUBROUTINE INTEGRATE( TIN, TOUT )

  USE GDATA

  IMPLICIT NONE !mz_rs_20030728

  ! TIN - Start Time          
  KPP_REAL TIN
  ! TOUT - End Time
  KPP_REAL TOUT

  CALL ROS2(VAR,TIN,DT)

  TIME = TOUT  

  RETURN
END SUBROUTINE INTEGRATE

! ****************************************************************************
! Rosenbrock integrator routine

SUBROUTINE ROS2(y,  t,  dtin)

  USE SPRSDATA

  ! Solve: d/dt c = f(t,c)
  !
  ! written by Edwin Spee, CWI, Amsterdam.
  ! email: Edwin.Spee@cwi.nl
  ! if non-existent, try http://edwin-spee.mypage.org/
  ! last update: July 28, 1997
  !
  ! on input:
  ! y      vector containing variable species at time=t
  ! fix    fixed species
  ! rconst reaction constants
  ! t      start time
  ! dtin   step size 
  !
  ! on output:
  ! y      vector containing variable species at time = tend
  ! other parameters are not changed
  !
  ! Uses datastructure and subroutine JACVAR_SP,DECOMP,F_VAR,SOLVE
  ! from KPP
  !
  ! Integration method for Ros2:
  ! C_{n+1} = C_n + 3/2 dt k_1 + 1/2 dt k_2
  ! k_1 = S f(t_n, C_n)
  ! k_2 = S [f(t_{n+1},C_n + dt k_1) - 2 k_1]
  !
  ! where g = 1.0 + sqrt(0.5),
  ! S = (I - g dt J ) ^ {-1}
  ! with J the Jacobian
  !
  ! mz_rs_20010423: real changed to KPP_REAL

  IMPLICIT NONE !mz_rs_20030728

  KPP_REAL y(nvar), rad(nrad), k1(nvar), k2(nvar) & ! mz_rs_20020623: rad added
    ,w1(nvar) &
    ,g,dtin,t &
    ,J_chem(LU_NONZERO_V+1) &
    ,sxmin
  INTEGER ier,it,nt,j,i

  g = 1.0 + SQRT(0.5)
  CALL JACVAR_SP(y,rad,J_chem)
  DO j=1,LU_NONZERO_V
    J_chem(j) = -g*dtin*J_chem(j)
  ENDDO
  !rvk optionally cut this out and replace it by directly addressed statements
  DO j=1,NVAR
    J_chem(lu_diag_v(j)) = J_chem(lu_diag_v(j)) + 1.0
  ENDDO

  !***********************************************************************
  !rvk call to old routine with indirect adressing 
  CALL DECOMP (NVAR, J_chem, ier)
  ! new calls with direct addressing (needs to be preprocessed by write_decomp2d.f)
  ! uncomment if wanted
  ! additional input: sxmin = sxmin = sqrt(smallest representable number > 0
  !     sxmin = 1.e-19 ! 4.e-154 for CRAY or double precision

  !     call DECOMP_D (J_chem,ier,sxmin)  ! box model call
  !     call DECOMP_2D (J_chem,ier,sxmin)  ! 2D call
  !***********************************************************************

  IF (ier.NE.0) THEN
    PRINT *,'IER <> 0, dtin=',dtin
    STOP
  ENDIF

  CALL F_VAR(y,rad,k1)

  CALL SOLVE (J_chem,k1)

  DO i = 1,nvar
    w1(i) = MAX(0.0_dp, y(i) + dtin * k1(i) ) ! mz_rs_20040830: dp added
  ENDDO

  CALL F_VAR(w1,rad,k2)

  DO i = 1,nvar
    k2(i) = k2(i) - 2.0 * k1(i)
  ENDDO
  CALL SOLVE (J_chem,k2)
  DO i = 1,nvar
    ! mz_rs_20040830: dp added to next line:
    y(i) = MAX(0.0_dp,y(i) + 1.5*dtin * k1(i) + 0.5 *dtin*k2(i) )
  ENDDO
END SUBROUTINE ROS2
