/***********************************************************************************
  KPP - symbolic chemistry Kinetics PreProcessor

  developed at Center for Global and Regional Atmospheric Research,
               University of Iowa

            by Valeriu Damian & Adrian Sandu 

***********************************************************************************/

#include "gdata.h"
#include "scan.h"
#include "y.tab.h"
#include <stdlib.h>
#include <string.h>
#include <math.h>

int AtomNr    = 0;
int SpeciesNr = 0;
int EqnNr     = 0;
int SpcNr     = 0;
int varNr     = 0;
int varactNr  = 0;
int radNr     = 0; 
int fixNr     = 0;
int varstNr   = 0;
int radstNr   = 0;
int fixstNr   = 0;

int generateSD = SD_ALL;

int initNr = -1;
int xNr = 0;
int yNr = 0;
int zNr = 0;

int falseSpcNr = 0;

ATOM_DEF AtomTable[ MAX_ATNR ];
SPECIES_DEF SpeciesTable[ MAX_SPECIES ];
CODE ReverseCode[ MAX_SPECIES ];
CODE Code[ MAX_SPECIES ];
KREACT kr[ MAX_EQN ];

float** cc;
float** cf;
float** cd;
int Reactive[ MAX_SPECIES ];
INLINE_KEY InlineKeys[] = { { F_DECL,     APPEND,  "F_DECL" },
                            { F_INIT,     APPEND,  "F_INIT" },
                            { F_DATA,     APPEND,  "F_DATA" },
                            { F_UTIL,     APPEND,  "F_UTIL" },
                            { F95_DECL,     APPEND,  "F95_DECL" },	/* F95 line */
                            { F95_INIT,     APPEND,  "F95_INIT" },
                            { F95_DATA,     APPEND,  "F95_DATA" },
                            { F95_UTIL,     APPEND,  "F95_UTIL" },
                            { F95_RCONST,   APPEND,  "F95_RCONST" }, /* mz_rs_20020426 */
                            { C_DECL,     APPEND,  "C_DECL" },
                            { C_INIT,     APPEND,  "C_INIT" },
                            { C_DATA,     APPEND,  "C_DATA" },
                            { C_UTIL,     APPEND,  "C_UTIL" },
		 	    { F_DECL_INT, REPLACE, "F_DECL_INT" },
                            { F_INIT_INT, REPLACE, "F_INIT_INT" },
                            { F_DATA_INT, REPLACE, "F_DATA_INT" },
                            { F_UTIL_INT, REPLACE, "F_UTIL_INT" },
		 	    { F95_DECL_INT, REPLACE, "F95_DECL_INT" },
                            { F95_INIT_INT, REPLACE, "F95_INIT_INT" },
                            { F95_DATA_INT, REPLACE, "F95_DATA_INT" },
                            { F95_UTIL_INT, REPLACE, "F95_UTIL_INT" },
                            { C_DECL_INT, REPLACE, "C_DECL_INT" },
                            { C_INIT_INT, REPLACE, "C_INIT_INT" },
                            { C_DATA_INT, REPLACE, "C_DATA_INT" },
                            { C_UTIL_INT, REPLACE, "C_UTIL_INT" }
		 	  };

int useAggregate   = 0;
int useJacobian    = 0;
int useAllJacobian = 0;
int useDouble      = 1;
int useReorder     = 1;
int useLang        = F_LANG;
int useJacSparse   = 0;

char integrator[ MAX_PATH ] = "rodas3";
char driver[ MAX_PATH ] = "matlab";
char runArgs[  MAX_PATH ] = "";

char varDefault[ MAX_IVAL ] = "1.E-8";
char radDefault[ MAX_IVAL ] = "1.E-8";
char fixDefault[ MAX_IVAL ] = "1.E-8";
double cfactor = 1.09E+10;

ATOM crtAtoms[ MAX_ATOMS ];
int crtAtomNr = 0;

char *fileList[ MAX_FILES ];
int fileNr = 0;

double Abs( double x ) 
{
  return x > 0 ? x : -x;
}

void DefineInitializeNbr( char *cmd )
{
int n;

  n = sscanf( cmd, "%d", &initNr);
  if( n != 1 )
    ScanError("Bad number of species to initialize <%s>", cmd);
}

void DefineXGrid( char *cmd )
{
int n;

  xNr = 1;
  n = sscanf( cmd, "%d", &xNr);
  if( n != 1 )
    ScanError("Bad X grid number <%s>", cmd);
}

void DefineYGrid( char *cmd )
{
int n;

  yNr = 1;
  n = sscanf( cmd, "%d", &yNr);
  if( n != 1 )
    ScanError("Bad Y grid number <%s>", cmd);
}

void DefineZGrid( char *cmd )
{
int n;

  zNr = 1;
  n = sscanf( cmd, "%d", &zNr);
  if( n != 1 )
    ScanError("Bad Z grid number <%s>", cmd);
}

void CmdFunction( char *cmd )
{
  if( EqNoCase( cmd, "AGGREGATE" ) ) {
    useAggregate = 1;
    return;
  }
  if( EqNoCase( cmd, "SPLIT" ) ) {
    useAggregate = 0;
    return;
  }
  ScanError("'%s': Unknown parameter for #FUNCTION", cmd );
}

void CmdJacobian( char *cmd )
{
  if( EqNoCase( cmd, "OFF" ) ) {
    useJacobian = 0;
    useAllJacobian = 0;
    useJacSparse = 0;
    return;
  }
  if( EqNoCase( cmd, "ON" ) ) {
    useJacobian = 1;
    useAllJacobian = 0;
    useJacSparse = 0;
    return;
  }
  if( EqNoCase( cmd, "ALL" ) ) {
    useJacobian = 1;
    useAllJacobian = 1;
    useJacSparse = 0;
    return;
  }
  if( EqNoCase( cmd, "SPARSE" ) ) {
    useJacobian = 1;
    useAllJacobian = 1;
    useJacSparse = 1;
    return;
  }
  ScanError("'%s': Unknown parameter for #JACOBIAN", cmd );
}

void CmdDouble( char *cmd )
{
  if( EqNoCase( cmd, "OFF" ) ) {
    useDouble = 0;
    return;
  }
  if( EqNoCase( cmd, "ON" ) ) {
    useDouble = 1;
    return;
  }
  ScanError("'%s': Unknown parameter for #DOUBLE", cmd );
}

void CmdReorder( char *cmd )
{
  if( EqNoCase( cmd, "OFF" ) ) {
    useReorder = 0;
    return;
  }
  if( EqNoCase( cmd, "ON" ) ) {
    useReorder = 1;
    return;
  }
  ScanError("'%s': Unknown parameter for #REORDER", cmd );
}

void CmdUse( char *cmd )
{
  if( EqNoCase( cmd, "FORTRAN95" ) ) { /* F95 lines */
    useLang = F95_LANG;
    return;
  }
  if( EqNoCase( cmd, "FORTRAN" ) ) {
    useLang = F_LANG;
    return;
  }
  if( EqNoCase( cmd, "C" ) ) {
    useLang = C_LANG;
    return;
  }
  ScanError("'%s': Unknown parameter for #USE", cmd );
}

void CmdIntegrator( char *cmd )
{
  strcpy( integrator, cmd );
}

void CmdDriver( char *cmd )
{
  strcpy( driver, cmd );
}

void CmdRun( char *cmd )
{
  strcpy( runArgs, cmd );
}

int FindAtom( char *atname )
{
int i;
  
  for( i=0; i<AtomNr; i++ )
    if( EqNoCase( AtomTable[ i ].name, atname ) ) {
      return i;   
    }
  return -1;      
}

void DeclareAtom( char *atname )
{
int code;

  code = FindAtom( atname );
  if ( code >= 0 ) {
    ScanError("Multiple declaration for atom %s.", atname );
    return;
  }
  if( AtomNr >= MAX_ATNR ) {
    Error("Too many atoms");
    return;
  }
  
  strcpy( AtomTable[ AtomNr ].name, atname );
  AtomTable[ AtomNr ].check = NO_CHECK;
  AtomTable[ AtomNr ].masscheck = 0;
  AtomNr++;
}

void SetAtomType( char *atname, int type )
{
int code;

  code = FindAtom( atname );
  if ( code < 0 ) {
    ScanError("Undefined atom %s.", atname );
    return;
  }
  AtomTable[ code ].check = type;
}

void CheckAll()
{
int i;
  
  for( i=0; i<AtomNr; i++ ) {
    if( AtomTable[ i ].check != CANCEL_CHECK )
      AtomTable[ i ].check = DO_CHECK;
  }   
  SetAtomType( "IGNORE", NO_CHECK );
}

void AddAtom( char *atname, char *nr )
{
int code;

  code = FindAtom( atname );
  if ( code < 0 ) {
    ScanError("Undefined atom %s.", atname );
    return;
  }
  crtAtoms[ crtAtomNr ].code = (unsigned char)code;
  crtAtoms[ crtAtomNr ].nr = (unsigned char)atoi(nr);
  crtAtomNr++;
}

int FindSpecies( char *spname )
{
int i;
  
  for( i=0; i<SpeciesNr; i++ )
    if( EqNoCase( SpeciesTable[ i ].name, spname ) ) {
      return i;   
    }
  for( i=0; i<2; i++ )
    if( EqNoCase( SpeciesTable[ MAX_SPECIES -1 - i ].name, spname ) ) {
      return MAX_SPECIES -1 - i;   
    }
  return -1;      
}

void StoreSpecies( int index, int type, char *spname )
{
int i;

  strcpy( SpeciesTable[ index ].name, spname );
  SpeciesTable[ index ].type = type; 
  *SpeciesTable[ index ].ival = '\0'; 
  SpeciesTable[ index ].lookat = 0;
  SpeciesTable[ index ].moni = 0;
  SpeciesTable[ index ].trans = 0;
  if( (SpeciesTable[ index ].nratoms == 0) || ( crtAtomNr > 0 ) ) {
    SpeciesTable[ index ].nratoms = crtAtomNr;
    for( i = 0; i < crtAtomNr; i++ )
      SpeciesTable[ index ].atoms[i] = crtAtoms[i];
  }
  crtAtomNr = 0;
}

void DeclareSpecies( int type, char *spname )
{
int code;

  code = FindSpecies( spname );
  if ( code >= 0 ) {
    ScanError("Multiple declaration for species %s.", spname );
    return;
  }
  if( SpeciesNr >= MAX_SPECIES ) {
    Error("Too many species");
    return;
  }
  StoreSpecies( SpeciesNr, type, spname );
  SpeciesNr++;
}

void SetSpcType( int type, char *spname )
{
int code;
int i;

  if( EqNoCase( spname, "VAR_SPEC" ) ) {
    for( i = 0; i < SpeciesNr; i++ ) 
      if( SpeciesTable[i].type == VAR_SPC ) 
        SpeciesTable[i].type = type;
    return;
  }
  if( EqNoCase( spname, "RAD_SPEC" ) ) {
    for( i = 0; i < SpeciesNr; i++ ) 
      if( SpeciesTable[i].type == RAD_SPC ) 
        SpeciesTable[i].type = type;
    return;
  }
  if( EqNoCase( spname, "FIX_SPEC" ) ) {
    for( i = 0; i < SpeciesNr; i++ ) 
      if( SpeciesTable[i].type == FIX_SPC ) 
        SpeciesTable[i].type = type;
    return;
  }
  if( EqNoCase( spname, "ALL_SPEC" ) ) {
    for( i = 0; i < SpeciesNr; i++ )
        SpeciesTable[i].type = type;
    return;
  }

  code = FindSpecies( spname );
  if ( code < 0 ) {
    ScanError("Undefined species %s.", spname );
    return;
  }
  SpeciesTable[ code ].type = type; 
}

void AssignInitialValue( char *spname , char *spval )
{
int code;  
double cf;
  
  if( EqNoCase( spname, "CFACTOR" ) ) {
    code = sscanf( spval, "%lg", &cf );
    if( code != 1 ) {
      ScanWarning("Invalid CFACTOR value: %s", spval);
      return;
    }
    cfactor = cf;
    return;
  }
  
  if( EqNoCase( spname, "VAR_SPEC" ) ) {
    strcpy( varDefault, spval );
    return;
  }
  
  if( EqNoCase( spname, "RAD_SPEC" ) ) {
    strcpy( radDefault, spval );
    return;
  }
  
  if( EqNoCase( spname, "FIX_SPEC" ) ) {
    strcpy( fixDefault, spval );
    return;
  }
  
  if( EqNoCase( spname, "ALL_SPEC" ) ) {
    strcpy( varDefault, spval );
    strcpy( radDefault, spval );
    strcpy( fixDefault, spval );
    return;
  }
  
  code = FindSpecies( spname );
  if ( code < 0 ) {
    ScanError("Undefined species %s.", spname );
    return; 
  }     
  strcpy( SpeciesTable[ code ].ival, spval );  
}

void StoreEquationRate( char *rate )
{
double f;
char buf[ MAX_K ];
int n;
KREACT *kreact;

  kreact = &kr[ EqnNr ]; 
  if( isPhoto ) {
    kreact->type = PHOTO;
    strcpy( kreact->val.st, rate );
    isPhoto = 0;
    return;
  }
  n = sscanf( rate, "%lf%s", &f, buf );
  if ( n == 1 ) {
    kreact->type = NUMBER;
    kreact->val.f = f;
    return;
  }
  kreact->type = EXPRESION;
  strcpy( kreact->val.st, rate );
  return;
}

void CheckEquation()
{
int i,j;
int equal, index;
double r1, r2;
float atcnt[ MAX_ATNR ];
int spc;
SPECIES_DEF *sp;
char errmsg[80];
int err;

  if( EqnNr >= MAX_EQN ) {
    Error("Too many equations");
    return;
  }
  
  for( i = 0; i < AtomNr; i++ )
    atcnt[i] = 0;
    
  for( spc = 0; spc < SpcNr; spc++ ) {
    sp = &SpeciesTable[ Code[spc] ];
    if( cc[spc][EqnNr] != 0 ) {
      for( i = 0; i < sp->nratoms; i++ )  
        atcnt[ sp->atoms[i].code ] += cc[spc][EqnNr] * sp->atoms[i].nr;
    }
    if( cd[spc][EqnNr] != 0 ) {
      for( i = 0; i < sp->nratoms; i++ ) 
        atcnt[ sp->atoms[i].code ] -= cd[spc][EqnNr] * sp->atoms[i].nr; 
    }
  } 
  
  *errmsg = 0;
  err = 0;
  
  for( i = 0; i < AtomNr; i++ ) {
    if ( Abs( atcnt[i] ) > 1e-5 ) {
      if ( AtomTable[i].check == CANCEL_CHECK ) {
        err = 0;
        break;
      }
      if ( AtomTable[i].check == NO_CHECK ) {
        continue;
      }
      if ( AtomTable[i].check == DO_CHECK ) {
        err = 1;
        sprintf(errmsg, "%s %s", errmsg, AtomTable[i].name );
        continue;
      }
    }
  } 
   
  if ( err ) 
    ScanWarning( "(eqn %d) Atom balance mismatch for:%s.", EqnNr+1, errmsg );    
        
  for( j = 0; j < SpcNr; j++ )
    if( cc[j][EqnNr] != 0 )
      { index = j; break; }
  for( i = 0; i < EqnNr; i++ ) {
    equal = 1;
    r1 = cc[index][EqnNr];
    r2 = cc[index][i];
    for( j = 0; j < SpcNr; j++ ) {
      if( r1 * cc[j][i] != r2 * cc[j][EqnNr] )
	{ equal = 0; break; }
      if( r1 * cd[j][i] != r2 * cd[j][EqnNr] )
	{ equal = 0; break; }
    }
    if ( equal ) {
      if( r1 == r2 )
        ScanError( "Duplicate equation: "
        	   " (eqn<%d> = eqn<%d> )", i+1, EqnNr+1 );
      else
	ScanError( "Linearly dependent equations: "
		   "( %.0f eqn<%d> = %.0f eqn<%d> )",
		   r1, i+1, r2, EqnNr+1 );
      break;
    }
  }
  EqnNr++;
}

void ProcessTerm( int side, char *sign, char *coef, char *spname  )
{
int code;  
CODE crtSpec;
double val;
char buf[40];


  code = FindSpecies( spname );
  if ( code < 0 ) {
    ScanError("Undefined species %s.", spname );
    return;
  }
  
  crtSpec = ReverseCode[ code ];

  if(EqNoCase(spname,"HV")) isPhoto = 1;

  if ( crtSpec == NO_CODE ) {
    if( MAX_SPECIES - code <= 2 ) falseSpcNr++;
    crtSpec = SpcNr++;
    Code[ crtSpec ] = code;
    ReverseCode[ code ] = crtSpec;
  }
  
  strcpy( buf, sign ); 
  strcat( buf, coef ); 
  sscanf( buf, "%lf", &val );

  switch( side ) {
    case LHS: cc[ crtSpec ][ EqnNr ] += val;
	      cf[ crtSpec ][ EqnNr ] -= val;
	      Reactive[ crtSpec ] = 1;
	      break;
    case RHS: cd[ crtSpec ][ EqnNr ] += val;
	      cf[ crtSpec ][ EqnNr ] += val;
	      break;
  }
}
           
void AddLumpSpecies( char *spname )
{
int code;  
  
  code = FindSpecies( spname );
  if ( code < 0 ) {
    ScanError("Undefined species %s.", spname );
    return;
  }

  /* ... */                

}

void CheckLump( char *spname )
{
int code;  
  
  code = FindSpecies( spname );
  if ( code < 0 ) {
    ScanError("Undefined species %s.", spname );
    return;
  }

  /* ... */                

}

void AddLookAt( char *spname )
{
int code;  
  
  code = FindSpecies( spname );
  if ( code < 0 ) {
    ScanError("Undefined species %s.", spname );
    return;
  }

  SpeciesTable[ code ].lookat = 1;   
}

void LookAtAll()
{
int i;

  for( i=0; i<SpeciesNr; i++ )
    SpeciesTable[ i ].lookat = 1;  
}

void AddMonitor( char *spname )
{
int code;  
  
  code = FindSpecies( spname );
  if ( code >= 0 ) {
    SpeciesTable[ code ].moni = 1;
    return;
  } 
  
  code = FindAtom( spname );
  if ( code >= 0 ) {
    AtomTable[ code ].masscheck = 1;
    return;
  }
    
  ScanError("Undefined species or atom %s.", spname );
}

void AddTransport( char *spname )
{
int code;  
  
  code = FindSpecies( spname );
  if ( code < 0 ) {
    ScanError("Undefined species %s.", spname );
    return;
  }

  SpeciesTable[ code ].trans = 1;   
}

void TransportAll()
{
int i;

  for( i=0; i<SpeciesNr; i++ )
    SpeciesTable[ i ].trans = 1;  
}

void AddUseFile( char *fname )
{
  fileList[fileNr] = (char*)malloc(strlen(fname)+1);
  strcpy(fileList[fileNr], fname);
  fileNr++;
}

char * AppendString( char * s1, char * s2, int * maxlen, int addlen ) 
{
char * tmp;

  *maxlen += addlen;

  if( !s1 ) {
    s1 = (char*)malloc( *maxlen );
    *s1 = 0;
  }    
  
  if( strlen( s1 ) + strlen( s2 ) >= *maxlen ) {
    s1 = (char*)realloc( (void*)s1, *maxlen );
  }
  strcat( s1, s2 ); 
  return s1;
}

char * ReplaceString( char * s1, char * s2, int * maxlen, int addlen ) 
{
char * tmp;

  if( s1 ) free(s1);

  *maxlen = strlen( s2 );
  s1 = (char*)malloc( 1+*maxlen );
  strcpy( s1, s2 );

  return s1;
}

void AddInlineCode( char * ctx, char * s )
{
ICODE * c;
int i, key, type;
  
  c = NULL;
  
  for( i = 0; i < INLINE_OPT; i++ ) 
    if( EqNoCase( ctx, InlineKeys[i].kname ) ) {
      key = InlineKeys[i].key;
      c = &InlineCode[key];
      type = InlineKeys[i].type;
      break;
    }
  if( !c ) {
    printf( "\n'%s': Unknown inline option (ignored)", ctx );
    return;
  }

  switch( type ) {
    case APPEND:  c->code = AppendString( c->code, s, &c->maxlen, MAX_INLINE );
                  break;
    case REPLACE: c->code = ReplaceString( c->code, s, &c->maxlen, MAX_INLINE );
                  break;
  }
}

void SparseData( char *cmd )
{
  if( EqNoCase( cmd, "OFF"      ) ) {generateSD = SD_OFF;      return;}
  if( EqNoCase( cmd, "ALL"      ) ) {generateSD = SD_ALL;      return;}
  if( EqNoCase( cmd, "LU_ROW"   ) ) {generateSD = SD_LU_ROW;   return;}
  if( EqNoCase( cmd, "JAC_ROW"  ) ) {generateSD = SD_JAC_ROW;  return;}
  if( EqNoCase( cmd, "LU_COL"   ) ) {generateSD = SD_LU_COL;   return;}
  if( EqNoCase( cmd, "JAC_COL"  ) ) {generateSD = SD_JAC_COL;  return;}
  if( EqNoCase( cmd, "LU_CROW"  ) ) {generateSD = SD_LU_CROW;  return;}
  if( EqNoCase( cmd, "JAC_CROW" ) ) {generateSD = SD_JAC_CROW; return;}
  if( EqNoCase( cmd, "LU_CCOL"  ) ) {generateSD = SD_LU_CCOL;  return;}
  if( EqNoCase( cmd, "JAC_CCOL" ) ) {generateSD = SD_JAC_CCOL; return;}

  ScanError("'%s': Unknown parameter for #SPARSEDATA", cmd );
}

int ParseEquationFile( char * filename )
{
int i,j;
int code;

  for( i = 0; i < MAX_SPECIES; i++ ) {
    ReverseCode[i] = NO_CODE;
    Reactive[i] = 0;
  }
  for( i = 0; i < MAX_SPECIES; i++ ) {
    for( j = 0; j < MAX_EQN; j++ ) {
      cc[i][j] = 0;
      cf[i][j] = 0;
      cd[i][j] = 0;
    }
  }
  for( i = 0; i < MAX_SPECIES; i++ ) {
    SpeciesTable[ i ].nratoms = 0;
  }
  
  for( i = 0; i < INLINE_OPT; i++ ) {
    InlineCode[i].code = NULL;
    InlineCode[i].maxlen = 0;
  }  
  
  EqnNr = 0;
  SpcNr = 0;
  
  DeclareAtom( "CANCEL" );   
  SetAtomType( "CANCEL", CANCEL_CHECK );
  DeclareAtom( "IGNORE" );   
  SetAtomType( "IGNORE", NO_CHECK );
  DeclareSpecies( DUMMY_SPC, "???" );
  StoreSpecies( MAX_SPECIES-1, DUMMY_SPC, "HV" );
  AddAtom( "CANCEL", "1" );
  StoreSpecies( MAX_SPECIES-2, DUMMY_SPC, "PROD" );

  code = Parser( filename );

  return code;
}

